#include "ServiceManager.hpp"

// C++ system level
#include <vector>
#include <algorithm>
// #include <cstring>     // memset, strncmp
// #include <cstdlib>     // TODO: is this for memcpy?
// #include <functional>

// ESP32 specific
#include "esp_log.h"
#include "freertos/FreeRTOS.h"
#include "freertos/task.h"

// project specific
#include <Types.hpp>
// #include "SiliconTorch/NVSExplorer.hpp"

// qthing stuff
// #include <qthing>


static const char* TAG = "ServiceManager";


namespace SiliconTorch {

  namespace ServiceManager {


    void registerSiliconTorchServices(ServiceManager* mgr);


    ServiceManager::ServiceManager() {

      registerSiliconTorchServices(this);

      registrationLocked = true;

      printServiceTable();

      ESP_LOGI(TAG, "Starting services…");
      u32 started = startEnabledServices();
      ESP_LOGI(TAG, "Starting services#[ %d ] done.", started);
    }

    std::vector<Service*> ServiceManager::getServicesInStartOrder() const {
      std::vector<Service*> services;

      for (const auto& [_, service] : serviceMap)
        services.push_back(service);

      std::sort(services.begin(), services.end(), [](const Service* x, const Service* y) {
        return x->getStartOrder() < y->getStartOrder();
      });

      return services;
    }

    u32 ServiceManager::startEnabledServices() {
      std::vector<Service*> services = getServicesInStartOrder();
      u32 started = 0;

      for (const auto& service : services) {
        if (service->isEnabled()) {
          service->start();
          started++;
        }
      }
      
      return started;
    }


    void ServiceManager::printServiceTable() const {

      // burn a few cycles because Espressivs logging is mehr :/
      // while (xTaskGetTickCount() * portTICK_PERIOD_MS < 1000) {
      //   vTaskDelay(42 / portTICK_PERIOD_MS);
      // }
      vTaskDelay(100 / portTICK_PERIOD_MS);

      // helper function, never used outside of this
      // TODO: should we move it to utils…?
      auto strmul = [](const str& target, u32 count) {
        str out = "";

        for (u32 n = 0; n < count; n++)
          out += target;

        return out;
      };


      u8 nameColSize = 4;  // size of "name"

      for (const auto& [name, _] : serviceMap)
        if (name.length() > nameColSize) nameColSize = name.length();

      nameColSize += 2;  // padding


      u8 startOrderColSize = 10 + 2;  // size of "StartOrder" + padding
      u8 enabledColSize    =  8 + 2;  // size of "Enabled " + padding (must be even)


      ESP_LOGI(TAG, "┌%s┬%s┬%s┐",
        strmul("─", nameColSize).c_str(),
        strmul("─", startOrderColSize).c_str(),
        strmul("─", enabledColSize).c_str()
      );

      ESP_LOGI(TAG, "│ Name %s│ StartOrder %s│ Enabled  %s│",
        strmul(" ", nameColSize - 4 - 2).c_str(),
        strmul(" ", startOrderColSize - 10 - 2).c_str(),
        strmul(" ", enabledColSize - 8 - 2).c_str()
      );

      ESP_LOGI(TAG, "╞%s╪%s╪%s╡",
        strmul("═", nameColSize).c_str(),
        strmul("═", startOrderColSize).c_str(),
        strmul("═", enabledColSize).c_str()
      );

      std::vector<Service*> services = getServicesInStartOrder();

      for (const auto& service : services) {
        u8 nameSize = service->getName().length();
        ESP_LOGI(TAG, "│ %s %s│ % 5d %s│%s%s%s│",
          service->getName().c_str(),
          strmul(" ", nameColSize - 2 - nameSize).c_str(),
          service->getStartOrder(),
          strmul(" ", startOrderColSize - 2 - 5).c_str(),
          strmul(" ", (enabledColSize - 2) / 2).c_str(),
          service->isEnabled() ? "✅" : "❌",
          strmul(" ", (enabledColSize - 2) / 2).c_str()
        );
      }

      ESP_LOGI(TAG, "└%s┴%s┴%s┘",
        strmul("─", nameColSize).c_str(),
        strmul("─", startOrderColSize).c_str(),
        strmul("─", enabledColSize).c_str()
      );

    }



    void ServiceManager::setServiceEnabled(const str& name, bool enabled) {
      if (serviceMap.count(name) > 0) {
        serviceMap.at(name)->setEnabled(enabled);
      } else {
        // Service not found
        // TODO: logging…?
      }
    }

    bool ServiceManager::getServiceEnabled(const str& name) const {
      if (serviceMap.count(name) > 0) {
        return serviceMap.at(name)->isEnabled();
      } else {
        // Service not found
        // TODO: logging…?
        return false;
      }
    }


    void ServiceManager::registerService(Service* s) {
      if (!registrationLocked) {
        s->init();
        s->postInit();

        serviceMap[s->getName()] = s;

        ESP_LOGI(TAG, "Registered service[ %s ]", s->getName().c_str());
      } else {
        ESP_LOGW(TAG, "Service registration is locked");
      }
    }



  }
}
